#
# Project
#
# Copyright (C) 2007  Windmill Media
# Copyright (c) 2006-2007 Lele Forzani <lele@windmill.it>
#
#
# = ICServiceBackend.rb
#
# <short descr>
#
# Author:: Lele Forzani <lele@windmill.it>
# Documentation::  Lele Forzani <lele@windmill.it>
# License::
#   This is NOT free software. License for using this software is granted
#   by written permission from Lele Forzani <lele@windmill.it>
# Revision:: $Id: ICServiceBackend.rb 1750 2007-10-04 13:36:34Z lele $
#
#
# == Description
#
# Implements common methods for services
#
# == Annotations
#
#
# == Todo

require 'bender/classes/ICBackend'
require 'bender/tools/Utilities'
require 'fileutils'
require 'open3'

require 'bender/tools/SystemInterface'

class ICServiceBackend < ICBackend
VARDIR = '/var/lib/bender/vhosts-docker'

attr_accessor :serviceName
attr_accessor :serviceConfig

attr_accessor :pluginList


def initialize(localID, configRoot)
	super

	if(defined?(@serviceName))
		getServiceConfig()
	end

	initPlugins(@serviceName)
end

#
# interface
#
def provisionService(args = {})
	verifyMandatoryPaths
  true
end

def updateService(args = {})
  return false unless(checkTriggers('update'))
  true
end

def maintainService(args = {})
  return false unless(checkTriggers('maintain'))
  true
end

def write(path, data)
	if File::exists?("#{path}.aclck")
		log.debug("  + should write '#{fileout}' but is locked")
	else
		if File::exists?(path)
			log.debug("  + backup previus configuration file '#{path}'")
			File.open("#{path}.bck", "w+") { |f|
				f.write(File.read(path))
			}
		end
		File.open(path, "w+") { |f|
			log.debug("  + writing file '#{path}'")
			f.write(data)
		}
	end
end

def loadConfigFile(fname)
  return [[],[]] unless File.exists?(fname)
  fstab = File.read(fname)
  a,b = fstab.split(/\# BENDER.*/)
  [(a||'').lines.map(&:strip), (b||'').lines.map(&:strip).reject(&:blank?)]
end

def writeConfigFile(fname, managed)
  manual, _ = loadConfigFile(fname)
  data = [manual, '# BENDER - do not edit below here!', managed, ''].flatten.join("\n")
  write(fname, data)
end

# ===========================================================================
protected

#
# retrieves configuration for this service from xml config
#
def getServiceConfig()

	DEBUG { log.debug("Retrieving configuration for backend #{self.class}")}
	xmlConfigFile = @configRoot + "/" + @serviceName + ".xml"
	#print("xmlConfFile: #{xmlConfigFile}\n")
	@serviceConfig = ConfigFile.new(xmlConfigFile)
 	@serviceConfig.no_hash_key_element_path = true

end

#
# checks if mandatory paths are present
#
def verifyMandatoryPaths

	basePath =  baseConfig.get_entry('/ICSystemConf/BaseDir')
	paths = serviceConfig.get_list(mySvcConfPath("MandatoryPaths/*"))

	paths.each { |pathEntry|
		details = serviceConfig.get_parameters(mySvcConfPath("MandatoryPaths/Path[@name='#{pathEntry[:name]}']/*"))
		path = basePath + "/" + details["path"]

		begin
			SystemInterface.verifyPath(path, true, details['mode'], details['uid'], details['gid'])
		rescue SystemCallError => e
			log.error("verifyMandatoryPaths: error in adding path '#{path}: #{e.message}'")
		end
	}
end

#
# executes plugin for the service
#
def execPlugins(service, mode='provision', env=nil)
	pluginList.each {|plugin|
		log.debug("Running Plugin: #{plugin.class}")
		plugin.send(mode, env)
	}

  execNativePlugins(service, mode)
end

def execNativePlugins(service, mode)
  pluginDir = File.join(bender_root, '/classes/Plugins/native/', service.capitalize)
  list = File.exists?(pluginDir) && File.directory?(pluginDir) ? Dir.entries(pluginDir).sort : []

  list.each { | pluginfile |
    next if(pluginfile.to_s =~ /^\./)
    native = pluginDir + '/' + pluginfile.to_s
    log.debug("+ Plugins: executing native plugin '#{native}'")
    res = `#{native} #{mode}`
    log.info(res)
  }
end

def initPlugins(service)
	pluginDir = File.join(bender_root, '/classes/Plugins/', service.underscore.camelize)
	list = File.exists?(pluginDir) && File.directory?(pluginDir) ? Dir.entries(pluginDir).sort : []

  @pluginList = Array.new

  list.each { | pluginfile |
		next if(pluginfile.to_s =~ /^\./)
		next unless(pluginfile.to_s =~ /\.rb$/)

		include = pluginDir + '/' + pluginfile.to_s
		log.debug("+ Plugins: including plugin '#{include}'")
		require(include)

		plugin = str2class($pluginClass.to_s)
		begin
			hnd = plugin.new(localID, configRoot)
			@pluginList << hnd
		rescue
			log.error("Error while instancing plugin: #{pluginfile}: #{$!}\n\t#{$!.backtrace.join("\n\t")}")

		end
	}

end

def checkTriggers(mode)
  if(mySvcConf("/Triggers/#{mode.capitalize}") == 'true')
    log.debug("- CheckTriggers: config says this mode requires triggers")

    trigflag = AcDirtyServices.find(:first, :conditions => [ "(host = ? OR host = 'any') AND service = ?", localID, @serviceName ])

    if(!trigflag || !trigflag.ts)
      timestamp = Time.at(0)
    else
      timestamp = trigflag.ts
    end

    trigstate = AcCleanServices.find(:first, :conditions => [ "host = ? AND service = ?", localID, @serviceName ])
    if(!trigstate)
      trigstate = AcCleanServices.new
      trigstate.host = localID
      trigstate.service = @serviceName
      state = Time.at(0)
    else
      state = trigstate.ts
    end

    log.debug("- CheckTriggers: state = #{state}, dirty timestamp = #{timestamp}")

    if(state < timestamp)
      trigstate.ts = trigflag.ts
      trigstate.save!
      log.info("- service has been triggered")
      return(true)
    end

    log.info("- service trigger state is clean")
    return(false)
  end
  true
end




# =====================
public

#
# accessor shortcut
#
def mySvcConfPath(path)
	return "/ICServiceConfig/" + @serviceName + "/" + path
end

#
# accessor shortcut
#
def mySvcConf(path, mandatory=true)
	svcConf = mySvcConfPath(path)

	res = serviceConfig[svcConf]

	if(res.nil? || res.empty?)
		log.warn("Retrieved empty configuration verb '#{svcConf}' from '#{serviceConfig.filename}', is missing.")
		if(mandatory)
			raise ArgumentError
		end
	end

	return res
end

def changed?(file)
  sumfile = File.join(VARDIR, file)
  dirname = File.dirname(sumfile)
  prevsum = if File.exists?(sumfile)
    File.read(sumfile).lines[0].strip
  end
  changed = true
  if prevsum
    sum = Digest::MD5.hexdigest(File.read(file))
    changed = prevsum != sum
  end
  changed
end

def checksum!(file)
  sumfile = File.join(VARDIR, file)
  dirname = File.dirname(sumfile)
  FileUtils.mkdir_p(dirname)
  File.open(sumfile, 'w+') do | f |
    f.write Digest::MD5.hexdigest(File.read(file))
  end
end



def run(cmd, env = {})
  success = true
  status = nil
  log.debug "  + command: '#{cmd}'"
  begin
    Open3.popen3(env, "#{cmd}") do |stdin, stdout, stderr, thread|
      stdin.close_write
      eof = false
      outbuf = ''
      errbuf = ''
      begin
        files = [stdout, stderr]

        until eof do
          ready = IO.select(files)

          if ready
            readable = ready[0]

            readable.each do |f|
              next if f.eof
              if f == stdout
                outbuf += f.read_nonblock(1024)
                lines = outbuf.split(/(\r\n|\r|\n)/)
                outbuf = lines.pop
                lines.each do | l |
                  next if l.strip == ''
                  log.debug("  # -> #{l}")
                end
              else
                errbuf += f.read_nonblock(1024)
                lines = errbuf.split(/(\r\n|\r|\n)/)
                errbuf = lines.pop
                lines.each do | l |
                  next if l.strip == ''
                  log.error("  # -> #{l}")
                end
              end
            end
          end
          eof = stdout.eof && stderr.eof
        end
        lines = outbuf.split(/(\r\n|\r|\n)/)
        lines.each do | l |
          next if l.strip == ''
          log.debug("  # -> #{l}")
        end
        lines = errbuf.split(/(\r\n|\r|\n)/)
        lines.each do | l |
          next if l.strip == ''
          log.error("  # -> #{l}")
        end
      end
      status = thread.value
    end
    status.success?
  rescue SystemCallError => ex
    log.error("  # -> #{ex}")
  end
end
end

# $URL: https://svn.intercom.it/svn/intercom/projects/Bender-NG/lib/classes/ICServiceBackend.rb $
# $Rev: 1750 $
# $Author: lele $
# $Date: 2007-10-04 15:36:34 +0200 (Thu, 04 Oct 2007) $
# $Id: ICServiceBackend.rb 1750 2007-10-04 13:36:34Z lele $
