#
# Project
#
# Copyright (C) 2007  Windmill Media
# Copyright (c) 2006-2007 Lele Forzani <lele@windmill.it>
#
#
# = filename.rb
#
# <short descr>
#
# Author:: Lele Forzani <lele@windmill.it>
# Documentation::  Lele Forzani <lele@windmill.it>
# License::
#   This is NOT free software. License for using this software is granted
#   by written permission from Lele Forzani <lele@windmill.it>
# Revision:: $Id: WebHosts.rb 698 2007-03-14 16:32:06Z lele $
#
#
# == Description
#
#
# == Annotations
#
#
# == Todo
#

require 'bender/tools/ConfigObject'
require 'kyotocabinet'
require 'uuidtools'

module Webserver
  class WebHost
    BENDER_NAMESPACE = UUIDTools::UUID.sha1_create(UUIDTools::UUID_URL_NAMESPACE, "https://www.intercom.it/ns/bender")
    class RecordNotFound < StandardError; end

    class << self
      def open(path)
        unless @database
          if File::exists?(path)
            db = KyotoCabinet::DB.new
            unless db.open(path, KyotoCabinet::DB::OREADER)
              raise "Cannot open database '#{path}': #{db.error}"
            end
            @database = db
          else
            raise "Cannot open database '#{path}': file not found"
          end
        end
      end

      def find(id)
        raise 'No database open' unless @database
        data = @database.get(id)
        unless data
          raise "Cannot read record #{id}: #{@database.error}"
        end
        self.new data
      end

      def find_all
        raise 'No database open' unless @database
        ret = []
        @database.each do |k, data|
          ret << self.new(data)
        end
        ret
      end

      def find_by_username(username)
        raise 'No database open' unless @database
        ret = nil
        @database.each do |k, data|
          record = self.new(data)
          if record.username == username
            ret = record
            break
          end
        end
        raise RecordNotFound.new("Vhost \"#{username}\" not found") unless ret
        ret
      end

      def find_all_active
        raise 'No database open' unless @database
        find_all.select(&:active?)
      end

      def find_active_by_username(username)
        raise 'No database open' unless @database
        ret = nil
        @database.each do |k, data|
          record = self.new(data)
          if record.username == username && record.active?
            ret = record
            break
          end
        end
        raise RecordNotFound.new("Vhost \"#{username}\" not found") unless ret
        ret
      end

      def find_all_not_active
        raise 'No database open' unless @database
        find_all.reject(&:active?)
      end

      def find_not_active_by_username(username)
        raise 'No database open' unless @database
        ret = nil
        @database.each do |k, data|
          record = self.new(data)
          if record.username == username && !record.active?
            ret = record
            break
          end
        end
        raise RecordNotFound.new("Vhost \"#{username}\" not found") unless ret
        ret
      end
    end

    def initialize(marshalled)
      @_data = Marshal.load marshalled
    end

    def active?
      @_data[:state][:state] == 'active'
    end

    def username
      @_data[:cfg][:username]
    end

    def data
      data = @_data[:cfg].dup
      normalize(data)
    end

    protected
    def normalize(data)
      #data = OpenStruct.new(data)
      ret = {
        service: {
          id: data[:id],
          active: self.active?,
          state: self.active?,
          login: data[:username],
          password: data[:password],
          name: "vhost-%s" % data[:username],
          servername: data[:names].first[:name],
          aliases: data[:names][1..-1].map{|d| d[:name]},
          created_at: data[:created_at],
          updated_at: data[:updated_at],
          legacy: false,
          environment: nil,
          docker_image: nil
        },
        quota: {
          quota_name: data[:quota]||data[:quota_name],
          enforce: data[:enforce_quota],
          quotas: {
            cpu: data[:quota_cpu],
            mem: data[:quota_mem],
            fs: data[:quota_fs],
          }
        },
        php: {
          enabled: data[:php_enabled],
          version: data[:php_version].to_s.downcase,
          instances: data[:php_num_instances] || 'AUTO',
          show_errors: truth(data[:php_show_errors]),
          memory_limit: data[:php_memory_limit] || 64,
          upload_max_filesize: data[:upload_max_size].to_i,
          additional: data[:php_additional] || '',
          extensions: (data[:php_extensions]||'').split(',').map(&:strip).reject(&:blank?),
          legacy: {
            writable_storage: (data[:php_legacy_writable_storage]||'').lines.map(&:strip).reject(&:blank?)
          }
        },
        ssl: {
          enabled: data[:ssl_enabled],
          allow_http: !data[:ssl_enabled] || data[:ssl_allow_http],
          letsencrypt: data[:ssl_enabled] && data[:ssl_letsencrypt],
          certs: (data[:ssl_certs] || []).map { | crt |
            {
              domains: crt[:domains].split(',').map(&:strip).reject(&:blank?),
              crt_id: UUIDTools::UUID.sha1_create(BENDER_NAMESPACE, ['ssl-cert', data[:username], crt[:id], crt[:domains], crt[:ssl_cert], crt[:ssl_key]].join).to_s,
              crt: crt[:ssl_cert_data],
              key: crt[:ssl_key_data],
            }
          }
        },
        http: {
          enabled: true,
          apache2: {
            compat: truth(data[:httpd_compat]),
            modules: (data[:httpd_modules]||'').split(',').map(&:strip).reject(&:blank?),
            global_options: data[:httpd_global_options].split(',').join(' '),
            allow_override: data[:httpd_allow_override].split(',').join(' '),
            additional: data[:httpd_additional] || '',
          }
        },
        proxy: {
          max_body_size: data[:upload_max_size].to_i
        },
        ftp: [
          {
            login: data[:username],
            password: data[:password],
            enabled: true,
            path: '/'
          }
        ],
        smtp: nil,
        databases: nil,
      }

      ret[:smtp] = if data[:php_enabled] || data[:is_docker]
        {
          smtp_server: nil,
          smtp_port: nil,
          smtp_tls: false,
          smtp_starttls: false,
          domain: nil,
          username: data[:username],
          password: UUIDTools::UUID.sha1_create(BENDER_NAMESPACE, ['smtp-login', data[:username], data[:password], 'smtp-login'].join).to_s
        }
      end

      ret[:databases] = (data[:services] || []).map do | svc |
        {
          type: svc[:service_type],
          name: svc[:name],
          database: svc[:name],
          login: svc[:username],
          password: svc[:password],
          enabled: !svc[:is_destroyed]
        }
      end

      if data[:is_docker].present?
        ret[:service][:docker_image] = data[:docker_image]
        ret[:php] = {enabled: false}
        ret[:http] = {enabled: false}
      end

      ret[:ftp] += data[:ftp_accesses].map do | f |
        {
          login: f[:username],
          password: f[:password],
          path: f[:path],
          enabled: f[:enabled]
        }
      end

      ret = ConfigObject.new(ret)
      ret.service.environment = Hash[data[:env_vars].map{|e| [e[:key], e[:value]]}]
      ret
    end

    def truth(v)
      (v||'').to_s.upcase == 'Y'
    end
  end
end
