require 'bender/classes/Plugin'
require 'bender/classes/Models/Webserver/FTPUser'
require 'securerandom'
require 'bender/tools/SftpGo'

$pluginClass = :Filesystems

class Filesystems < ICPlugin
  def initialize(*args)
    super
  end

  def provision(args)
    host = args[:host]
    service = args[:service]
    env = args[:env]

    setupXFSProject(host, service, env)
    setupFilesystem(host, service, env)
  end

  def update(args)
    host = args[:host]
    service = args[:service]
    env = args[:env]

    setupXFSProject(host, service, env)
    setupFilesystem(host, service, env)
  end

  def maintain(args)
    host = args[:host]
    service = args[:service]
    env = args[:env]
  end

  def deactivate(args)
    host = args[:host]
    service = args[:service]
    env = args[:env]

    init_sftpgo
    disableNFSExport(host, service, env)
    removeFTPLogin(host, service, env)
  end

  def activate(args)
    host = args[:host]
    service = args[:service]
    env = args[:env]

    init_sftpgo
    enableNFSExport(host, service, env)
    addFTPLogin(host, service, env)
  end

  def prepare(args)
    service = args[:service]
    action = args[:action]
    env = args[:env]
  end

  def cleanup(args)
    service = args[:service]
    action = args[:action]
    env = args[:env]
  end

  protected

  def init_sftpgo
    url = baseConfig.get_string('/ICSystemConf/sftpgo/url')
    @sftpgo = SftpGo.new url
  end

  def setupFilesystem(host, service, env)
    storagePath = service.host_storagePath(host)
    mntPath = service.containerVHostRoot(host)

    if host.service.docker_image.present?
      storagePath = File.join(storagePath, service.mySvcConf("/ServicePaths/UserSpace/DocRootPath"))
      mntPath = File.join(mntPath, service.mySvcConf("/ServicePaths/UserSpace/DocRootPath"))
    end

    enabled = nfsEnabled?(host, service, env)

    env[:volumes] = {
      storage: {
        volume: {
          device: ":%s" % [storagePath],
          options: ["addr=%s"%service.mySvcConf("/VolumeStorage/Server"), service.mySvcConf("/VolumeStorage/Options")].compact.join(','),
          type: 'nfs'
        },
        mount: mntPath,
        read_only: false
      },
      #logs: {
      #  volume: {
      #    device: ":%s" % [service.host_logPath(host)],
      #    options: ["addr=%s"%service.mySvcConf("/VolumeStorage/Server"), service.mySvcConf("/VolumeStorage/Options"), 'ro'].compact.join(','),
      #    type: 'nfs'
      #  },
      #  mount: File.join(service.containerVHostRoot(host),
      #    service.mySvcConf("/ContainerPaths/AccessLogPath")),
      #  read_only: true
      #}
    }
  end

  def setupXFSProject(host, service, env)
    return unless supports_quota?
    projects_file = baseConfig.get_string('/ICSystemConf/XFS/ProjectsFile')
    prjid_file = baseConfig.get_string('/ICSystemConf/XFS/ProjectIdsFile')

    storagePath = service.host_storagePath(host)
    log.debug '  + setup quota'

    idx = service.svcUID(host) * 50 + host.service.id
    prjid  = "%s:%d" % [host.service.login, idx]
    prj    = "%d:%s" % [idx, storagePath]

    _, projects = service.loadConfigFile(projects_file)
    unless projects.include?(prj)
      projects << prj
      service.writeConfigFile(projects_file, projects)
    end

    _, projids = service.loadConfigFile(prjid_file)
    unless projids.include?(prjid)
      projids << prjid
      service.writeConfigFile(prjid_file, projids)
    end

    quota = if host.quota.enforce
      q = if host.quota.quota_name
        service.mySvcConf("/Quotas/Quota[@name='#{host.quota.quota_name}']/Filesystem").to_i
      elsif host.quota.quotas.fs
        host.quota.quotas.fs.try(:to_i)
      end
      if q
        q += (q/10).to_i
        "%sM" % q
      end
    end
    unless quota
      quota = service.mySvcConf("/ConfigDefaults/UndefinedQuota")
      log.debug "  + quota undefined, use default of #{q}" unless host.quota.enforce
      log.debug "  + do not enforce quota, use default of #{q}" if host.quota.enforce
    end

    mountpoint = getMntPointFor(storagePath)
    log.debug "  + quota size: #{quota}"
    service.run "xfs_quota -x -c 'project -s #{host.service.login}' #{mountpoint}"
    service.run "xfs_quota -x -c 'limit -p bhard=#{quota} #{host.service.login}' #{mountpoint}"
  end


  def enableNFSExport(host, service, env)
    path = service.host_storagePath(host)
    log.debug "  + enable nfs export of #{path}"

    export = "%s %s(%s,fsid=%d)" % [
      path,
      service.mySvcConf('/NFS/Data/ACL'),
      service.mySvcConf('/NFS/Data/Options'),
      host.service.id
    ]

    addNFSExport(path, export, service, env)

    #path = service.host_logPath(host)
    #log.debug "  + enable nfs export of #{path}"

    #export = "%s %s(%s,fsid=%d)" % [
    #  path,
    #  service.mySvcConf('/NFS/Logs/ACL'),
    #  service.mySvcConf('/NFS/Logs/Options'),
    #  host.service.id
    #]

    #addNFSExport(path, export, service, env)
  end

  def disableNFSExport(host, service, env)
    path = service.host_storagePath(host)
    log.debug '  + disable nfs export of #{path}'

    removeNFSExport(path, service, env)

    #path = service.host_logPath(host)
    #log.debug '  + disable nfs export of #{path}'

    #removeNFSExport(path, service, env)
  end

  def addNFSExport(path, export, service, env)
    exports_file = baseConfig.get_string('/ICSystemConf/NFS/ExportsFile')
    _, exports = service.loadConfigFile(exports_file)
    unless exports.include?(export)
      exports.reject!{|e| e =~ /\A#{path}\ /}
      exports << export
      service.writeConfigFile(exports_file, exports)
      log.debug '  + reload nfs server'
      service.run service.mySvcConf('/Commands/NFSReload')
    end
  end

  def removeNFSExport(path, service, env)
    exports_file = baseConfig.get_string('/ICSystemConf/NFS/ExportsFile')
    _, exports = service.loadConfigFile(exports_file)
    exports.reject!{|e| e =~ /\A#{path}\ /}
    service.writeConfigFile(exports_file, exports)
    log.debug '  + reload nfs server'
    service.run service.mySvcConf('/Commands/NFSReload')
  end

  def nfsEnabled?(host, service, env)
    path = service.host_storagePath(host)
    _, exports = service.loadConfigFile('/etc/exports')
    !exports.find{|e| e =~ /\A#{path}\ /}.blank?
  end

  def supports_quota?
    true # TODO detect xfs
  end

  def getMntPointFor(path)
    `stat -c '%m' "#{path}"`.strip
  end

  def removeFTPLogin(host, service, env)
    @sftpgo.delete_group host.service.name
  end

  def addFTPLogin(host, service, env)
    group = @sftpgo.get_group host.service.name
    group ||= @sftpgo.create_group SftpGo::Group.new('name' => host.service.name)

    active = host.ftp.map do | ftp |
      user = @sftpgo.get_user ftp.login
      user ||= SftpGo::User.new
      user.username = ftp.login
      user.password = ftp.password
      user.uid = service.svcUID(host)
      user.gid = service.mySvcConf('/SystemUsers/DefaultGroupId')
      user.home_dir = File.join(service.host_storagePath(host), ftp.path)
      user.groups = [{name: group.name, type: 3}]
      user.active = ftp.enabled ? 1 : 0

      if user.id
        @sftpgo.update_user user
      else
        @sftpgo.create_user user
      end
    end
  end
end
