require 'bender/classes/ICServiceBackend'
require 'bender/classes/Models/Webserver/WebHosts'
require 'fileutils'
require 'mail'
require 'mysql2'

class ICMysql < ICServiceBackend

  SERVICE = 'mysql'

  def initialize(localID, configRoot)

    @serviceName = SERVICE
    super

    Webserver::WebHost.open(baseConfig.get_string('/ICSystemConf/DatabasePath'))

    @masterConn = baseConfig.get_string_array('/ICSystemConf/VhostsDatabase')[0]
    @master = Mysql2::Client.new(@masterConn)

    @slaveConn = baseConfig.get_string_array('/ICSystemConf/VhostsSlaveDatabase')[0]
    @slave = Mysql2::Client.new(@slaveConn)
  end

  def provisionService(args)
    return unless super
  end

  def updateService(args)
    return unless super
  end

  def maintainService(args)
    return unless super
    dumpDatabases(args) if checkSlaveReplica
  end

  protected

  def dumpDatabases(args)
    log.debug " + STOP SLAVE SQL_THREAD;"
    @slave.query('STOP SLAVE SQL_THREAD;')

    dumpsDir = mySvcConf('DumpsDir')
    FileUtils.mkdir_p(dumpsDir)
    eachHost(args) do | host |
      log.info " + Backup databases for #{host.service.name}"
      host.databases.each do | db |
        baseFname = File.join(dumpsDir, "#{host.service.name}-#{db.database}-")
        fname = "#{baseFname}#{Time.new.strftime("%Y%m%d")}.sql.gz"
        tempfile = "#{dumpsDir}.tmp.gz"
        cmd = "mysqldump --opt -u%s -p'%s' -h %s -P%d %s | gzip > %s" % [
          @slaveConn['username'],
          @slaveConn['password'],
          @slaveConn['host'],
          @slaveConn['port'],
          db.database,
          tempfile
        ]
        log.debug "  + Backup database #{db.database} to temp file"
        if run cmd
          log.debug "  + rename #{tempfile} to #{fname}"
          FileUtils.mv(tempfile, fname)
          old = Dir["#{baseFname}*"].sort
          while old.length > mySvcConf('Retention').to_i
            del = old.shift
            log.debug "  + remove old backup #{del}"
            File.unlink(del)
          end
        else
          File.unlink(tempfile) if File.exists(tempfile)
        end
      end if host.databases
    end
  ensure
    log.debug " + START SLAVE;"
    @slave.query('START SLAVE;')
  end

  def eachHost(args = {}, &proc)
    webHosts = Webserver::WebHost.find_all_active if args[:service].blank?
    webHosts = [Webserver::WebHost.find_active_by_username(args[:service])] unless args[:service].blank?

    webHosts.each do | h |
      proc.call(h.data)
    end
  end

  def checkSlaveReplica
    log.debug ' + Checking slave sync status'

    q = "SHOW MASTER STATUS;"
    log.debug "  + #{q}"
    master = @master.query(q)
    data = master.first
    mLog, mLogPos = data['File'], data['Position']

    q = "SHOW SLAVE STATUS;"
    log.debug "  + #{q}"
    slave = @slave.query(q)
    data = slave.first
    sLog, sLogPos = data['Master_Log_File'], data['Read_Master_Log_Pos']

    log.debug "  + master: #{mLog} #{mLogPos}"
    log.debug "  + slave:  #{sLog} #{sLogPos}"

    if (sLog == mLog && sLogPos >= mLogPos) || (sLog > mLog)
      log.info ' + Slave is syncronized with Master'
      true
    else
      if sLog == mLog
        log.error " + Slave is behind Master!"
        log.error "  + master: #{mLog} #{mLogPos}"
        log.error "  + slave:  #{sLog} #{sLogPos}"
        log.error "  + Dump databases anyway"
        notifyReplicaError(mLog, mLogPos, sLog, sLogPos, true)
        true
      else
        log.error " + Slave is VERY BEHIND Master!"
        log.error "  + master: #{mLog} #{mLogPos}"
        log.error "  + slave:  #{sLog} #{sLogPos}"
        log.error "  + WILL NOT DUMP ANY DATABASE!"
        notifyReplicaError(mLog, mLogPos, sLog, sLogPos, false)
        false
      end
    end
  end

  def notifyReplicaError(mLog, mLogPos, sLog, sLogPos, dump)

    body = if dump
      %Q(
Host: #{@localID}
Mysql replica problem!

Slave is behind Master
master: #{mLog} #{mLogPos}
slave:  #{sLog} #{sLogPos}

Will perform database backups.
)
    else
      %Q(
Host: #{@localID}
MYSQL REPLICA PROBLEM!

Slave is dangerously behind Master
master: #{mLog} #{mLogPos}
slave:  #{sLog} #{sLogPos}

NO DATABASE BACKUP PERFORMED!
)
    end

    to = baseConfig.get_string_array('/ICSystemConf/Notifications/To/Addr').map{|a| a['Addr']}
    mail = Mail.new do | m |
      m.from    baseConfig['/ICSystemConf/Notifications/From']
      m.to      to
      m.subject "[#{localID}] MYSQL REPLICA PROBLEM!"
      m.body    body
    end
    mail.delivery_method :sendmail

    if mail.deliver
      log.warn " * notification sent!"
    else
      log.error " * cannot send notification!"
    end
  end
end





