

require 'pathname'
require 'uri'
require 'pp'
require 'ipaddr'

require 'bender/tools/Validator'



# Collection of high-level validation helpers for O/R mapping objects
# (Derivatives of ActiveRecord::Base
# 
# All methods are built around AcriveRecord::Base#validates_each
class ActiveRecord::Base
  
  protected
  
  # Check whether the given attributes contain valid identifiers
  def self.validates_identifier(*attr_names)
    conf = { :message => 'is not a valid identifier', :on => :save, :allow_nil => false }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    conf[:with] = /\A[a-zA-Z_]\w*\z/
    
    validates_format_of(attr_names, conf)
  end # validates_identifier
  
  
  # Check whether the given attributes contain  well-formed email adresses
  def self.validates_email(*attr_names)
    conf = { :message => " is not a well-formed email adress", :on => :save, :allow_nil => false }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    validates_each(attr_names, conf) do |record, attr_name, value|
      # strip comments
      val = value.to_s
      val.gsub!(/\([^()]*\)/, '')
      val.gsub!(/"[^"]*"/, '')
      val.sub!(/^.*<(.*)>\s*$/, '\\1')
      
      # strip leading/trailing whitespace
      val.strip!
      
      record.errors.add(attr_name, conf[:message]) unless val =~ EMAIL_REGEX
    end
  end # validates_email
  
  
  # Check whether the given attributes contain well-formed hostnames or IP addresses
  # (ipv4 only for now)
  def self.validates_host_or_ip(*attr_names)
    conf = { :message => 'is not a well-formed hostname or IPv4 address', :on => :save, :allow_nil => false }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    validates_each(attr_names, conf) do |record, attr_name, value|
      # well-formed hostname? => ok
      next if(value =~ Validator::ValueChecker::HOSTNAME_REGEX)

      # otherwise check if it's an IP address
      ip = IPAddr.new(value, Socket::AF_INET) rescue record.errors.add(attr_name, conf[:message])
    end
  end # validates_host_or_ip
  
  
  # Check whether the given attributes contain well-formed hostnames
  def self.validates_hostname(*attr_names)
    conf = { :message => 'is not a well-formed hostname', :on => :save, :allow_nil => false, :allow_unqualified => true }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    if(conf[:allow_unqualified])
      conf[:with] = Validator::ValueChecker::HOSTNAME_REGEX
    else
      conf[:with] = Validator::ValueChecker::Q_HOSTNAME_REGEX
    end
    
    validates_format_of(attr_names, conf)
  end # validates_hostname
  
  
  # Check whether the given attributes contain IP addresses (IPv4 only for now)
  def self.validates_ipv4(*attr_names)
    conf = { :message => 'is not a well-formed IPv4 address', :on => :save, :allow_nil => false }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    validates_each(attr_names, conf) do |record, attr_name, value|
      ip = IPAddr.new(value, Socket::AF_INET) rescue record.errors.add(attr_name, conf[:message])
    end
  end # validates_ipv4
  
  
  # Check whether the given attributes contain well-formed filesystem paths
  #
  # Nonstandard Options:
  #  * :kind : One of "abs" (absolute paths only), "rel" (relative paths only) or anything else (don't care) [any]
  #  * :allow_up : (boolean) Whether the paths may contain ".." parts [false]
  #  * :depth :    (range) The number of elements the path may contain. Use together with allow_up=false! [no limit]
  def self.validates_path(*attr_names)
    conf = { :message => " is not a well-formed path", :on => :save, :allow_nil => false, :allow_up => false }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    validates_each(attr_names, conf) do |record, attr_name, value|
      p = Pathname.new(value.to_s).cleanpath
      
      if((conf[:kind] == "abs") && p.relative?)
        record.errors.add(attr_name, "has to be an absolute path")
        next
      elsif((conf[:kind] == "rel") && p.absolute?)
        record.errors.add(attr_name, "has to be a relative path")
        next
      end
      
      count = 0
      p.each_filename { |part|
        if(!conf[:allow_up] && (part == '..'))
          record.errors.add(attr_name, "may not contain '..'")
          next
        end
        count += 1
      }
      
      if(conf[:depth] && (!conf[:depth].include?(count)))
        record.errors.add(attr_name, "must have a depth in (#{conf[:depth].first}..#{conf[:depth].last})")
        next
      end
    end
  end # validates_path
  
  
  # Check if the given attributes contain well-formed URLs
  #
  # Nonstandard Options:
  #  * :schemes (array) List of allowed URL schemes ['http', 'https', 'ftp']
  def self.validates_url(*attr_names)
    conf = { :message => "is not a well-formed URL", :on => :save, :allow_nil => false, :schemes => ['http', 'https', 'ftp'] }
    conf.update(attr_names.pop) if attr_names.last.is_a?(Hash)
    
    validates_each(attr_names, conf) do |record, attr_name, value|
      tmp = URI::parse(value) rescue record.errors.add(attr_name, conf[:message])
      unless(!tmp.kind_of?(URI) || conf[:schemes].include?(tmp.scheme))
        record.errors.add(attr_name, "has an invalid URL scheme (allowed are #{conf[:schemes].join(',')})")
      end
    end        
  end
  

  
  private
  
  # FIXME: This is duplicated in ValueChecker!
  EIGHTBIT_REGEX_S = '(0*(\d{1,2}|1\d{2}|2[0-4]\d|25[0-5]))'
  IPV4_REGEX_S     = "(#{EIGHTBIT_REGEX_S}\\.){3}#{EIGHTBIT_REGEX_S}"
  HOSTNAME_REGEX_S = '([\w-]+\.)*([a-zA-Z]{2,})'
  
  EMAIL_REGEX = /\A[^@\s]+@[\w-]+(\.[\w-]+)*?(\.\w\w+)\z/
  HOST_REGEX  = Regexp.new("(\\A#{IPV4_REGEX_S}\\z)|(\\A#{HOSTNAME_REGEX_S}\\z)")
  
end # class ActiveRecord::Base
