require 'net/http'
require 'json'
require 'ostruct'

class SftpGo
  class ApiError < StandardError
    attr_reader :error, :message
    def initialize(error)
      super [error['message'], error['error']].reject(&:blank?).join(': ')
      @error = error['error']
      @message = error['message']
    end

    def not_found?
      self.error == 'Not Found'
    end
  end

  class Record < OpenStruct
    def self.fields
      []
    end

    def initialize(data = {})
      defaults = self.class.fields.zip([nil] * self.class.fields.count).to_h
      raise ApiError.new(data) if (data.keys & ['error', 'message']).any?
      super defaults.merge(data)
    end

    def as_json(**params)
      data = super **params
      data['table']
    end
  end

  class ApiKey < Record
    def self.fields
      ['access_token', 'expires_at']
    end

    def expires_at
      Time.parse(@table[:expires_at]) if @table[:expires_at]
    end

    def expired?
      !expires_at || expires_at <= Time.new
    end
  end

  class Group < Record
    def self.fields
      ["id", "name", "description", "created_at", "updated_at", "user_settings", "virtual_folders", "users", "admins"]
    end

    def to_json
      data = as_json except: [:id, :created_at, :updated_at, :users, :admins]
      data.delete('description') unless data['description']
      data.delete('user_settings') unless data['user_settings']
      data.delete('virtual_folders') unless data['virtual_folders']
      JSON.dump(data)
    end
  end

  class User < Record
    def self.fields
      [
        "id", "status", "username", "email", "description", "expiration_date", "password", "public_keys", "has_password",
        "home_dir", "virtual_folders", "uid", "gid", "max_sessions", "quota_size", "quota_files", "permissions", "used_quota_size",
        "used_quota_files", "last_quota_update", "upload_bandwidth", "download_bandwidth", "upload_data_transfer", "download_data_transfer",
        "total_data_transfer", "used_upload_data_transfer", "used_download_data_transfer", "created_at", "updated_at",
        "last_login", "first_download", "first_upload", "last_password_change", "filters", "filesystem", "additional_info",
        "groups", "oidc_custom_fields", "role"
      ]
    end

    def status
      @table['status'] ||= 1
    end

    def uid
      @table['uid'] ||= 2000
    end

    def gid
      @table['gid'] ||= 1000
    end

    def permissions
      @table['permission'] ||= {
        "/": [
          "*"
        ],
      }
    end

    def to_json
      data = as_json except: [:id, :created_at, :updated_at, :users, :admins]
      data['status'] ||= status
      data['permissions'] ||= permissions
      data.keys.each do | k |
        data.delete(k) unless data[k].present?
      end
      data['uid'] = data['uid'].to_i if data['uid']
      data['gid'] = data['gid'].to_i if data['gid']
      JSON.dump(data)
    end
  end

  def initialize(url)
    @sftpgo_url = url
    login
  end

  def get_group(name)
    begin
      Group.new(get("/api/v2/groups/#{name}"))
    rescue ApiError => err
      if err.not_found?
        debug "Group \"#{name}\" not found"
        nil
      else
        raise err
      end
    end
  end

  def create_group(group)
    Group.new post("/api/v2/groups", group)
  end

  def update_group(group)
    Group.new put("/api/v2/groups/#{group.name}", group)
  end

  def delete_group(group)
    name = group.name if group.is_a?(Group)
    name = group if group.is_a?(String)
    group = get_group(name)
    if group
      if group.users
        group.users.each do | u |
          delete_user u
        end
      end
      response = delete("/api/v2/groups/#{name}")
      raise ApiError.new(response) if response.keys.include? 'error'
    end
    true
  end

  def get_user(name)
    begin
      User.new(get("/api/v2/users/#{name}"))
    rescue ApiError => err
      if err.not_found?
        debug "User \"#{name}\" not found"
        nil
      else
        raise err
      end
    end
  end

  def create_user(user)
    User.new post("/api/v2/users", user)
  end

  def update_user(user)
    response = put("/api/v2/users/#{user.username}", user)
    raise ApiError.new(response) if response.keys.include? 'error'
    get_user(user.username)
  end

  def delete_user(user)
    username = user.username if user.is_a?(User)
    username = user if user.is_a?(String)
    user = get_user username
    if user
      response = delete("/api/v2/users/#{username}")
      raise ApiError.new(response) if response.keys.include? 'error'
    end
    true
  end

  protected

  def debug(message)
    $log.try(:debug, message)
  end

  def info(message)
    $log.try(:info, message)
  end

  def error(message)
    $log.try(:error, message)
  end

  def api_key
    @api_key = nil if @api_key && @api_key.expired?
    unless @api_key
      @api_key = login
    end
    @api_key.try(:access_token)
  end

  def api_uri
    @api_uri ||= URI(@sftpgo_url)
  end

  def http
    @http ||= begin
      http = Net::HTTP.new(api_uri.host, api_uri.port)
      logger = Class.new do
        def <<(msg)
          $log.debug(msg) if @log
        end
      end.new
      http.set_debug_output logger
      http
    end
  end

  def login
    req = Net::HTTP::Get.new('/api/v2/token')
    req.basic_auth api_uri.user, api_uri.password

    http.start do | http |
      response = http.request req
      ApiKey.new(JSON.parse(response.body))
    end
  end

  def delete(path, params = {})
    req = Net::HTTP::Delete.new(path)
    req['Accept'] = 'text/json'
    req['Authorization'] = "Bearer #{api_key}"

    http.start do | http |
      response = http.request req
      JSON.parse(response.body)
    end
  end

  def get(path, params = {})
    req = Net::HTTP::Get.new(path)
    req['Accept'] = 'text/json'
    req['Authorization'] = "Bearer #{api_key}"

    http.start do | http |
      response = http.request req
      JSON.parse(response.body)
    end
  end

  def post(path, data, params = {})
    req = Net::HTTP::Post.new(path)
    req['Accept'] = 'text/json'
    req['Conten-Type'] = 'text/json'
    req['Authorization'] = "Bearer #{api_key}"
    req.body = data.try(:to_json) || data.to_s

    http.start do | http |
      response = http.request req
      JSON.parse(response.body)
    end
  end

  def put(path, data, params = {})
    req = Net::HTTP::Put.new(path)
    req['Accept'] = 'text/json'
    req['Conten-Type'] = 'text/json'
    req['Authorization'] = "Bearer #{api_key}"
    req.body = data.try(:to_json) || data.to_s

    http.start do | http |
      response = http.request req
      JSON.parse(response.body)
    end
  end
end
