module Validator
  
  # Collection of methods for checking values for validity.
  #
  # All methods will throw an +ArgumentErrror+ if they find the value to
  # be invalid and silently return if it's valid.
  module ValueChecker
    
    #include Contractor
    
    # Might be useful for other code
    HOSTNAME_REGEX    = /\A([\w-]+\.)*([a-zA-Z]{2,})\z/
    Q_HOSTNAME_REGEX  = /\A([\w-]+\.)+([a-zA-Z]{2,})\z/
    EMAIL_REGEX       = /\A[^@\s]+@[\w-]+(\.[\w-]+)*?(\.\w\w+)\z/
    
    
    protected
    
    
    # Check for a boolean condition
    def checkTrue(aCondition, aMsg = nil)
      aMsg = "'True' value expected, but condition evaluated to 'false'" unless aMsg
      raise(ArgumentError, aMsg, caller(1)) unless aCondition
    end # checkTrue()
    
    
    # Check whether aString is a valid Identifier
    def checkIdentifier(aString, aMsg = nil)
      checkKindOf(String, aString)
      
      aMsg = "'#{aString}' is not a valid identifier" unless aMsg
      checkTrue((aString =~ /\A[a-zA-Z_]\w*\z/), aMsg)
    end # checkIdentifier
    
    
    # Check whether aValue == anExpected
    def checkEqual(anExpected, aValue, aMsg = nil)
      aMsg = "'#{anExpected}' expected, but got '#{aValue}'" unless aMsg
      checkTrue((aValue == anExpected),aMsg)
    end # checkEqual
    
    
    # Check whether aValue.instance_of? aClass
    def checkInstanceOf(aClass, aValue, aMsg = nil)
      requireKindOf(Class, aClass)
      
      aMsg = "Has to be of type '#{aClass.name}'" unless aMsg
      checkTrue((aValue.instance_of? aClass),aMsg)
    end # checkInstanceOf
    
    
    # Check whether aValue.kind_of? aClass
    def checkKindOf(aClass, aValue, aMsg = nil)
      requireKindOf(Class, aClass)
      
      aMsg = "Has to be of a subtype of '#{aClass.name}'" unless aMsg
      checkTrue((aValue.kind_of? aClass), aMsg)
    end # checkKindOf
    
    
    # Check whether aValue == nil
    def checkNil(aValue, aMsg = nil)
      aMsg = "Has to be nil (must not contain any value)" unless aMsg
      checkTrue((aValue == nil),aMsg)
    end # checkNil
    
    
    # Check whether aString =~ aPattern
    def checkMatch(aPattern, aString, aMsg = nil)
      checkKindOf(String, aString)
      
      if(aPattern.kind_of? Regexp)
        pat = aPattern
      else
        pat = Regexp.new(aPattern.to_s)
      end
      
      aMsg = "Has to match regular expression '#{aPattern.to_s}'" unless aMsg
      checkTrue((aString =~ pat), aMsg)
    end # checkMatch
    
    
    # Check whether aString !~ aPattern
    def checkNotMatch(aPattern, aString, aMsg = nil)
      checkKindOf(String, aString)
      
      if(aPattern.kind_of? String)
        pat = Regexp.new(aPattern)
      else
        pat = aPattern
      end
      
      aMsg = "May not match regular expression '#{aPattern.to_s}'" unless aMsg
      checkTrue((aString !~ pat), aMsg)
    end # checkNotMatch
    
    
    # Check whether aValue <= aMaxSize
    def checkMaxSize(aMaxSize, aValue, aMsg = nil)
      requireKindOf(Integer, aMaxSize)
      requireTrue(aMaxSize >= 0)
      
      aMsg = "Size(Length) has to be at most #{aMaxSize} but is #{aValue.size}" unless aMsg
      checkTrue((aValue.size <= aMaxSize), aMsg)
    end # checkMaxSize
    
    
    # Check whether aValue >= aMinSize
    def checkMinSize(aMinSize, aValue, aMsg = nil)
      requireKindOf(Integer, aMinSize)
      requireTrue(aMinSize >= 0)
      
      aMsg = "Size(Length) has to be at least #{aMinSize} but is #{aValue.size}" unless aMsg
      checkTrue((aValue.size >= aMinSize), aMsg)
    end # checkMinSize
    
    
    # Check whether aSizeRange === aValue
    def checkSize(aSizeRange, aValue, aMsg = nil)
      requireKindOf(Range, aSizeRange)
      
      aMsg = "Size(Length) has to be in range (#{aSizeRange.first}..#{aSizeRange.last}) but is #{aValue.size}" unless aMsg
      checkTrue((aSizeRange === aValue.size), aMsg)
    end
    
    
    # Check whether aValue != nil
    def checkNotNil(aValue, aMsg = nil)
      aMsg = "May not be nil (has to have a value)" unless aMsg
      checkTrue((aValue != nil),aMsg)
    end # checkNotNil
    
    
    # Check whether anEnumerable.include?(aValue)
    def checkInSet(anEnumerable, aValue, aMsg = nil)
      requireKindOf(Enumerable, anEnumerable)
      
      if(aMsg.nil?)
        if(anEnumerable.kind_of?(Array))
          aMsg = 'Has to be one of "' + anEnumerable.join('", "') + '"'
        elsif(anEnumerable.kind_of?(Range))
          aMsg = "Has to be in (#{anEnumerable.first}..#{anEnumerable.last}), but is #{aValue}"
        else
          aMsg = "Has to be in a given set"
        end
      end
      
      checkTrue(anEnumerable.include?(aValue), aMsg)
    end # checkInSet
    
    
    # Check whether aValue is a well-formed email adress
    def checkEmail(aValue, aMsg = nil)
      aMsg = "Not a well-formed email adress: '#{aValue}'" unless aMsg
      
      # strip comments
      val = aValue
      val.gsub!(/\([^()]*\)/, '')
      val.gsub!(/"[^"]*"/, '')
      val.sub!(/^.*<(.*)>\s*$/, '\\1')
      
      # strip leading/trailing whitespace
      val.strip!
      
      ##puts "after preprocessing: '#{val}'\n"
      # check for basic form
      checkTrue((val =~ EMAIL_REGEX), aMsg)
    end # checkEmail
    
    
    # Check whether aValue is a well-formed hostname or IP address
    # (IPv4 only for now)
    def checkHostOrIp(value, aMsg = nil)
      aMsg = "Not a well-formed hostname or ipv4 address: '#{value}'" unless aMsg
      
      # check for hostname
      begin
        checkMatch(HOSTNAME_REGEX, value, aMsg)
      rescue ArgumentError => e
        # otherwise check if it's an IP address
        ip = IPAddr.new(value, Socket::AF_INET) 
      end
    end # checkHostOrIp
    
  end # module ValueChecker
  
end # module Validator